/*  fgwirc - fungw based minimal irc client
    Copyright (C) 2019  Tibor 'Igor2' Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    Project page and contact: http://repo.hu/projects/fgwirc
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <termios.h>
#include <unistd.h>
#include <sockev/sockev.h>
#include <sockev/sockev_astdio.h>

#include "action.h"
#include "edit.h"
#include "window.h"

#include "input.h"

sockev_t s_input;
char *input_key_binding[IMOD_max][INPU_MAX_KEYCODE];
int input_meta = 0;

input_mod_t input_name2mod(const char *name)
{
	if (isdigit(*name)) {
		int n = atoi(name);
		if ((n < 0) || (n >= IMOD_max))
			return IMOD_invalid;
		return n;
	}

	if (strcmp(name, "none") == 0) return IMOD_NONE;
	if (strcmp(name, "ctrl") == 0) return IMOD_CTRL;
	if (strcmp(name, "alt") == 0) return IMOD_ALT;
	if (strcmp(name, "pad") == 0) return IMOD_PAD;
	if (strcmp(name, "pad2") == 0) return IMOD_PAD2;
	if (strcmp(name, "pad3") == 0) return IMOD_PAD3;
	if (strcmp(name, "func") == 0) return IMOD_FUNC;
	if (strcmp(name, "meta") == 0) return IMOD_META;

	return IMOD_invalid;
}

static void input_enter(void)
{
	struct termios t;

	if (tcgetattr(0, &t) < 0)
		return;

	t.c_lflag &= ~ICANON;
	t.c_lflag &= ~ECHO;
	t.c_cc[VMIN] = 0;
	t.c_cc[VTIME] = 10;

	tcsetattr(0, TCSANOW, &t);
}

static void input_leave(void)
{
	struct termios t;

	if (tcgetattr(0, &t) < 0)
		return;

	t.c_lflag |= ICANON;
	t.c_lflag |= ECHO;

	tcsetattr(0, TCSANOW, &t);
}

static void ev_input_close(sockev_t id)
{
	input_leave();
}

static int input_key(input_mod_t modifier, int keycode)
{
/*	fprintf(stderr, "-> [%d:%d]\n", modifier, keycode);*/
	if ((modifier < 0) || (modifier >= IMOD_max))
		return -1;
	if ((keycode < 0) || (keycode > INPU_MAX_KEYCODE))
		return -1;

	if (input_key_binding[modifier][keycode] != NULL) {
		action_str(input_key_binding[modifier][keycode]);
		return 1;
	}

	/* fallback: let the editor try to handle it */
	return edit_key(modifier, keycode);
}

typedef enum {
	IST_PLAIN,
	IST_ESC,
	IST_ESC79,
	IST_ESC91,
	IST_ESC91_49,
	IST_ESC91_50,
	IST_ESC91_IGNORE_126
} ist_t;

ist_t ist;

static int input_state_machine(int chr)
{
	switch(ist) {
		case IST_ESC91_IGNORE_126:
			ist = IST_PLAIN;
			if (chr == 126)
				return 0;
			/* fall through */
		case IST_PLAIN:
			if (chr == 27) {
				ist = IST_ESC;
				return 0;
			}
			if ((chr == '\n') || (chr == '\r') || (chr == 8))
				return input_key(0, chr);
			if ((chr >= 32) && (chr < 128)) {
				if (input_meta) {
					input_meta = 0;
					return input_key(IMOD_META, chr);
				}
				return input_key(IMOD_NONE, chr);
			}
			if (chr < 32)
				return input_key(IMOD_CTRL, chr);
			return -1;
		case IST_ESC:
			if (chr == 91) {
				ist = IST_ESC91;
				return 0;
			}
			if (chr == 79) {
				ist = IST_ESC79;
				return 0;
			}
			ist = IST_PLAIN;
			if (chr < 32)
				return input_key(IMOD_CTRL, chr);
			return input_key(IMOD_ALT, chr);
		case IST_ESC91:
			if (chr == 49) {
				ist = IST_ESC91_49;
				return 0;
			}
			if (chr == 50) {
				ist = IST_ESC91_50;
				return 0;
			}
			if (chr <= 55)
				ist = IST_ESC91_IGNORE_126;
			else
				ist = IST_PLAIN;
			return input_key(IMOD_PAD, chr);
		case IST_ESC91_49:
			if (chr == 126) {
				ist = IST_PLAIN;
				return input_key(IMOD_PAD, 49);
			}
			ist = IST_ESC91_IGNORE_126;
			return input_key(IMOD_PAD2, chr);
		case IST_ESC91_50:
			if (chr == 126) {
				ist = IST_PLAIN;
				return input_key(IMOD_PAD, 50);
			}
			ist = IST_ESC91_IGNORE_126;
			return input_key(IMOD_PAD3, chr);
		case IST_ESC79:
			ist = IST_PLAIN;
			return input_key(IMOD_FUNC, chr);
	}
	return -1;
}

static void ev_input_read(sockev_t id, char *buff, int len)
{
	int n;
/*	fprintf(stderr, "key: %d:", len);*/
	for(n = 0; n < len; n++) {
/*		fprintf(stderr, " %d", buff[n]);*/
		input_state_machine(tcurses_getch(term, buff[n]));
	}
/*	fprintf(stderr, "\n");*/
}

static sockev_hooks_t ev_input = {NULL, ev_input_close, NULL, NULL, NULL, ev_input_read, NULL};

void input_init(void)
{
	input_enter();
	s_input = sockev_new(&ev_input, 0, NULL);
	sockev_astdio(&s_input);
}
